import numpy as np
import pandas as pd
from sklearn.decomposition import PCA
from sklearn.preprocessing import MinMaxScaler

def normalizer_or_matrix(file, biological_activity):
	# Biological activity:
	if len(biological_activity) == 1:
		bio_array = (file[biological_activity].to_numpy(copy=True)).reshape(-1,1) #Transforming the list of biological activities into an array of shape -1:1.
	else:
		bio_array = file[biological_activity].to_numpy(copy=True) # Transforming the list of biological activities into an array.
	scaler = MinMaxScaler() # Definition of the normalization parameter.
	bio_array_normalized = scaler.fit_transform(bio_array) # Normalization.
	
	# Physicochemical descriptors:
	descriptors_physicochemical = ['NumHAcceptors', 'NumHDonors', 'ExactMolWt', 'NumRotatableBonds', 'FractionCSP3', 'TPSA', 'LogP_WildmanCrippen']
	PhCh_array = file[descriptors_physicochemical].to_numpy(copy=True) # Insertion of physicochemical properties into an array.
	scaler = MinMaxScaler() # Definition of the normalization parameter.
	PhCh_array_normalized = scaler.fit_transform(PhCh_array) # Normalization.

	# Fingerprint:
	FP_array = np.array(list(file['AtomPairs'])) # Insertion of fingerprints into an array.
	return bio_array_normalized, PhCh_array_normalized, FP_array

def pca_maker(file, nPCS, svd_parameter):
	# (if/elif:) Setting parameters to skip PCA: If the number of properties is less than or equal to 3 and if it is less than or equal to the number of principal components desired.
	if (nPCS >= file.shape[1]):
		pca_props = file
	elif (file.shape[1] <= 3):
		pca_props = file
	else: # If it did not meet any exclusion criteria, proceed with the analysis.
		pca = PCA(n_components=nPCS, svd_solver=svd_parameter)
		pca_props = pca.fit_transform(file)
	return pca_props

def organize_hca(file, HCAdict, ident): # Add the cluster identification to the spreadsheet.
	if ident == 'all':
		file['Cluster_General'] = pd.Series(HCAdict)
		return file
	elif ident == 'bio':
		file['Cluster_Biological'] = pd.Series(HCAdict)
		return file
	elif ident == 'PhCh':
		file['Cluster_Physicochemical'] = pd.Series(HCAdict)
		return file
	else:
		file['Cluster_Structural'] = pd.Series(HCAdict)
		return file

def organize_for_kmodes(file): # Create a matrix with cluster identifications for each of the three domains, in order to prepare for Kmodes.
	kmodes_columns = file[['Cluster_Biological', 'Cluster_Physicochemical', 'Cluster_Structural']]
	dict_cluster_bio = {}
	dict_cluster_PhCh = {}
	dict_cluster_FP = {}
	for i, a, b, c in zip(kmodes_columns.index, kmodes_columns.Cluster_Biological, kmodes_columns.Cluster_Physicochemical, kmodes_columns.Cluster_Structural):
		dict_cluster_bio[i] = 'cluster '+str(a)
		dict_cluster_PhCh[i] = 'cluster '+str(b)
		dict_cluster_FP[i] = 'cluster '+str(c)

	kmodes_new_columns = pd.DataFrame()
	kmodes_new_columns['Cluster_Biological'] = pd.Series(dict_cluster_bio)
	kmodes_new_columns['Cluster_Physicochemical'] = pd.Series(dict_cluster_PhCh)
	kmodes_new_columns['Cluster_Structural'] = pd.Series(dict_cluster_FP)
	kmodes_matrix = np.array(kmodes_new_columns)
	return kmodes_matrix


import numpy as np
from matplotlib import pyplot as plt
from matplotlib import colors as colrs
import scipy.cluster.hierarchy as sch
from sklearn.cluster import AgglomerativeClustering
from kmodes import kmodes

def numerical_number_of_clusters(distances_list):
	# Elbow method using the Euclidean distance:
	shorted_list = distances_list[0:21] # Create a list of distances from the first 21 clusters.
	x1, y1 = 1, shorted_list[0] # Define the x-axis coordinates of the first and last points on the line.
	x2, y2 = len(shorted_list)-1, shorted_list[len(shorted_list)-1] # Define the y-axis coordinates of the first and last points on the line.
	distances = []
	for i in range(len(shorted_list)):
		x0 = i+1 # Coordinate on the x-axis of the point to calculate the distance to the line.
		y0 = shorted_list[i] # Coordinate on the y-axis of the point to calculate the distance to the line.
		numerator = abs((y2-y1)*x0 - (x2-x1)*y0 + x2*y1 - y2*x1)
		denominator = ((y2 - y1)**2 + (x2 - x1)**2)**(1/2)
		distances.append(numerator/denominator) # Distance between a point and a line.
	return (distances.index(max(distances)) + 1)

def categorical_number_of_clusters(file):
	# Elbow method using the clustering cost:
	listClusCosts = []
	for n in range(2,21): # Calculate the cost for each n cluster.
		km = kmodes.KModes(n_clusters=n, verbose=0, n_init=20, init='Cao', max_iter=300)
		km.fit(X=file)
		listClusCosts.append(km.cost_) # Costs = Clustering cost, defined as the sum distance of all points to their respective cluster centroids.
	x1, y1 = 2, listClusCosts[0] # Define the x-axis coordinates of the first and last points on the line.
	x2, y2 = 20, listClusCosts[len(listClusCosts)-1] # Define the y-axis coordinates of the first and last points on the line.

	distances = []
	for i in range(len(listClusCosts)):
		x0 = i+2 # Coordinate on the x-axis of the point to calculate the distance to the line.
		y0 = listClusCosts[i] # Coordinate on the y-axis of the point to calculate the distance to the line.
		numerator = abs((y2-y1)*x0 - (x2-x1)*y0 + x2*y1 - y2*x1)
		denominator = ((y2 - y1)**2 + (x2 - x1)**2)**(1/2)
		distances.append(numerator/denominator) # Distance between a point and a line.
	return (distances.index(max(distances)) + 2)

def optimal_threshold(file, ident, directoryFileOutput, extension_type):
	clustering = AgglomerativeClustering(compute_distances=True, linkage='complete',affinity='euclidean').fit(file) # Perform HCA.
	distances_list = (sorted(list(clustering.distances_))) # Extract the calculated Euclidean distance between clusters.
	distances_list.reverse() # Sort the list in descending order. 

	number_of_clusters = numerical_number_of_clusters(distances_list) # Calculate the number of clusters for the HCA.
	CutOff = distances_list[number_of_clusters - 2] # Get the cutoff threshold for the dendrogram.

	if ident == 'bio':
		name_euc_dist_file = directoryFileOutput+'\\Images\\Distance_HCA_Images\\Dist_HCA_biological.'+extension_type # Define the name, directory and extension of the Euclidean distance graph.
		title_distances = 'Plot of Euclidean Distances from Biological HCA'
		print('Number of clusters:')
		print('Biological:', number_of_clusters) # Print the number of clusters for the biological domain.
		plot_euclidean_distance(distances_list, name_euc_dist_file, title_distances) # Plot the graph of Euclidean distances.
	elif ident == 'PhCh':
		name_euc_dist_file = directoryFileOutput+'\\Images\\Distance_HCA_Images\\Dist_HCA_physicochemical.'+extension_type # Define the name, directory and extension of the Euclidean distance graph.
		title_distances = 'Plot of Euclidean Distances from Physicochemical HCA'
		print('Physicochemical:', number_of_clusters) # Print the number of clusters for the physicochemical domain.
		plot_euclidean_distance(distances_list, name_euc_dist_file, title_distances) # Plot the graph of Euclidean distances.
	else:
		name_euc_dist_file = directoryFileOutput+'\\Images\\Distance_HCA_Images\\Dist_HCA_structural.'+extension_type # Define the name, directory and extension of the Euclidean distance graph.
		title_distances = 'Plot of Euclidean Distances from Structural HCA'
		print('Structural:', number_of_clusters) # Print the number of clusters for the structural domain.
		plot_euclidean_distance(distances_list, name_euc_dist_file, title_distances) # Plot the graph of Euclidean distances.

	return CutOff, number_of_clusters

def plot_euclidean_distance(distances_list,name_euc_dist_file,title_distances): # Plot the graph of Euclidean distances.
	plt.plot(list(range(1, len(distances_list)+1)), distances_list, '-o', color = 'black', markersize=4, linewidth=1, markeredgewidth=1) # Initial setting of graph values and parameters.
	plt.ylabel('Euclidean distance') # Set the y-axis label.
	original_ticks = list((plt.xticks())[0]) # Get the x-axis ticks.
	increase_rate = original_ticks[2] # Get the x-axis increment.
	list_x = list(range(1, len(distances_list)+1,int(increase_rate/2))) # Create a list to define the new xticks that correspond to double the points compared to the previous one.
	plt.xticks(list_x, list_x, fontsize=8) # Define the new x-ticks and set the font size.
	plt.title(title_distances) # Define the graph title.
	plt.savefig(name_euc_dist_file) # Save figure.
	plt.close() # Close plot.

def hca_clusters(file, labels, ident, directoryFileOutput, extension_type):
	# Get the threshold value (CutOff) and the number of clusters:
	CutOff, number_of_clusters = optimal_threshold(file, ident, directoryFileOutput, extension_type)
	linkage = sch.linkage(file, method = 'complete', optimal_ordering=False)

	# Get the cluster labels:
	leaves_cluster = sch.fcluster(linkage, t=number_of_clusters, criterion='maxclust')
	cluster_dict = dict(zip(labels, list(leaves_cluster)))

	return leaves_cluster, cluster_dict, linkage, CutOff

def hca_plot(linkage, labels, leaves_cluster, CutOff, ident, directoryFileOutput, extension_type, dendrogram_xfont_size, index_of_test_molecules):
	# Create an arrow for molecules in test set:
	labels_arrow = labels.copy()
	for i in range(0,len(labels_arrow)):
		if labels_arrow[i] in index_of_test_molecules:
			labels_arrow[i] = ('→ '+labels_arrow[i])

	# Working with colors:
	hexRGB = ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd', '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf', '#691605', '#b3f084', '#ffee00', '#00ff6e', '#0400ff', '#c4c4c0', '#ff00fb', '#290e0e', '#12290e', '#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd', '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf', '#691605', '#b3f084', '#ffee00', '#00ff6e', '#0400ff', '#c4c4c0', '#ff00fb', '#290e0e', '#12290e'] # Definition of which colors (colorblind friendly) will be used for each cluster (1-19).
	sch.set_link_color_palette(hexRGB) # Set hexRGB as the dendrogram colors.

	# Initial setting of HCA dendrograms:
	plt.figure(figsize=(20, 10), dpi= 300) # Set the image space.
	dendrogram = sch.dendrogram(linkage, color_threshold=CutOff, above_threshold_color='k', labels=labels_arrow, get_leaves=True)

	# Set graphs font size for both axis and adjust the plot in image space:
	plt.yticks(fontsize=12)
	plt.xticks(fontsize=dendrogram_xfont_size)
	plt.ylabel('Euclidean distance',fontsize=12)
	plt.subplots_adjust(left=0.06 ,right=0.96)

	# Set new scale for y axis (which is 2x the number of points in the previous scale):
	b = [float(i) for i in plt.yticks()[0]]
	scale = b[1]-b[0]
	new_scale = scale/2
	new_arange = np.arange(min(b)-scale/20,max(b)+new_scale,new_scale)
	plt.yticks(new_arange)

	# Assign the legend to each cluster:
	set__leaves_cluster = sorted(list(set(leaves_cluster)))
	set_leaves_cluster = []
	for i in set__leaves_cluster:
		if i < 10:
			set_leaves_cluster.append('cluster 0'+str(i))
		else:
			set_leaves_cluster.append('cluster '+str(i))
	plt.legend(set_leaves_cluster)

	# Add ticks to x-axis:
	plt.tick_params(axis='x', which='both', bottom=True, top=False, labelbottom=True)

	# Get the color for each cluster:
	color_list = list(dendrogram['leaves_color_list'])

	# Apply cluster color to x-labels:
	zip_label_color = list(zip(plt.gca().get_xticklabels(), color_list))
	for i,a in zip_label_color:
		i.set_color(a)

	# Define the title of the graph and saves the figure:
	if ident == 'bio':
		plt.title('HCA of biological activity')
		plt.savefig(directoryFileOutput+'\\Images\\Dendrogram_HCA_Images\\HCA_biological.'+extension_type)
		plt.close()
	elif ident == 'PhCh':
		plt.title('HCA of physicochemical properties')
		plt.savefig(directoryFileOutput+'\\Images\\Dendrogram_HCA_Images\\HCA_physicochemical.'+extension_type)
		plt.close()
	else:
		plt.title('HCA of AtomPairs fingerprint')
		plt.savefig(directoryFileOutput+'\\Images\\Dendrogram_HCA_Images\\HCA_structural.'+extension_type)
		plt.close()

def kmodes_clusters(file, names):
	# Calculate the number of clusters for the Kmodes:
	n_clusters = categorical_number_of_clusters(file)
	print('General:', n_clusters) # Print the number of clusters.

	# Run Kmodes with the calculated number of clusters:
	km_test = kmodes.KModes(n_clusters=n_clusters, n_init=20, init='Cao', max_iter=300) # Parameter setting.
	clusters = km_test.fit_predict(file) # Run Kmodes for the dataset and return an array with cluster labels for each molecule.
	clusters_list = list(clusters) # Transform np.array into a list.

	# Create a dictionary with the pair "molecule name": "General cluster ID":
	HCA_General_dict = {}
	for i, a in zip(names, clusters_list):
		HCA_General_dict[i] = int(a+1)
	return HCA_General_dict


import pandas as pd
import numpy as np
from sklearn.model_selection import train_test_split
from matplotlib import pyplot as plt
from matplotlib import colors as colrs

def split_train_test_sets(file, training_percent, test_percent):
	training, test = train_test_split(file, test_size=test_percent, train_size=training_percent, stratify=file['Cluster_General']) # Split the set of molecules into training set and test set, preserving the cluster diversity of the total set according to the stratification of the general clustering.
	test_molecules = [i for i in test.index] # Extraction of the name of the molecules included in the test set.
	dict_training = {'set':'training'}
	dict_test = {'set': 'test'}
	training = training.assign(**dict_training) # Assign the training label to the molecules in the training set.
	test = test.assign(**dict_test) # Assign the test label to the molecules in the test set.
	dataset = pd.concat([training, test]) # Group all molecules into a single dataframe.
	return dataset, test_molecules

def freq_clusters(file, directoryFileOutput, extension_type, Cluster_Domain, barplot_xfont_size):
	# Creating lists with the cluster IDs of each molecule:
	Total_count = []
	Training_count = []
	Test_count = []
	
	for index, row in file.iterrows():
		if row['set'] == 'training':
			Training_count.append(row[Cluster_Domain])
			Total_count.append(row[Cluster_Domain])
		if row['set'] == 'test':
			Test_count.append(row[Cluster_Domain])
			Total_count.append(row[Cluster_Domain])
	min_value = min(Total_count)
	max_value = max(Total_count)
	
	# Frequency calculations for each cluster in each set saved in dictionaries:
	clusters_total = {}
	clusters_training = {}
	clusters_test = {}
	for i in range(min_value, (max_value+1)):
		if i <= 9:
			clusters_total[('cluster 0'+str(i))] = ((Total_count.count(i))*100)/len(Total_count)
			clusters_training[('cluster 0'+str(i))] = ((Training_count.count(i))*100)/len(Training_count)
			clusters_test[('cluster 0'+str(i))] = ((Test_count.count(i))*100)/len(Test_count)
		else:
			clusters_total[('cluster '+str(i))] = ((Total_count.count(i))*100)/len(Total_count)
			clusters_training[('cluster '+str(i))] = ((Training_count.count(i))*100)/len(Training_count)
			clusters_test[('cluster '+str(i))] = ((Test_count.count(i))*100)/len(Test_count)

	# Transform the dictionaries into a list of tuples and sort the list based on the first element of the tuple (the cluster ID):
	list_total = sorted(list(clusters_total.items()), key=lambda x:x[0])
	list_training = sorted(list(clusters_training.items()), key=lambda x:x[0])
	list_test = sorted(list(clusters_test.items()), key=lambda x:x[0])

	# Separate the cluster IDs from their frequencies:
	label_total = [list_total[i][0] for i in range(0,len(list_total))]
	percentages_total = [list_total[i][1] for i in range(0,len(list_total))]
	percentages_training = [list_total[i][1] for i in range(0,len(list_training))]
	percentages_test = [list_test[i][1] for i in range(0,len(list_test))]

	# Create an "np.arange" for each set:
	Total_ar = np.arange(len(list_total))
	Training_ar = np.arange(len(list_training))
	Test_ar = np.arange(len(list_test))
	
	# Bar Plot (matplotlib)
	plt.rcParams.update({'font.size':12}) # Set the font size in all fields of the chart.
	bar_width = 0.25
	fig,ax = plt.subplots(figsize = (10,6)) # Set the figure size.
	group1 = ax.bar(Total_ar-bar_width, percentages_total, bar_width, label = 'Total', color='black', edgecolor = 'black') # Bar plot for the total set.
	group2 = ax.bar(Training_ar, percentages_training, bar_width, label = 'Training set', color='dimgrey', edgecolor = 'black') # Bar plot for the training set.
	group3 = ax.bar(Test_ar+bar_width, percentages_test, bar_width, label = 'Test set', color= 'whitesmoke', edgecolor = 'black') # Bar plot for the test set.
	ax.legend() # Set the plot legend.
	ax.set_ylabel('Frequency (%)') # Set the y-axis label.
	ax.set_ylim([0,100]) # Set the y-axis range.
	ax.set_xticks(Total_ar) # Set the x-axis increment.
	ax.set_xticklabels(label_total, fontsize=barplot_xfont_size) # Set the font size and label of x-axis.

	if Cluster_Domain == 'Cluster_Biological':
		ax.set_title('Distribution of dataset: Biological Clustering')
		plt.savefig(directoryFileOutput+'\\Images\\Distribution\\Biological.'+extension_type)
	elif Cluster_Domain == 'Cluster_Structural':
		ax.set_title('Distribution of dataset: Structural Clustering')
		plt.savefig(directoryFileOutput+'\\Images\\Distribution\\Structural.'+extension_type)
	elif Cluster_Domain == 'Cluster_Physicochemical':
		ax.set_title('Distribution of dataset: Physicochemical Clustering')
		plt.savefig(directoryFileOutput+'\\Images\\Distribution\\Physicochemical.'+extension_type)
	elif Cluster_Domain == 'Cluster_General':
		ax.set_title('Distribution of dataset: General Clustering')
		plt.savefig(directoryFileOutput+'\\Images\\Distribution\\General.'+extension_type)
	plt.close()
	return clusters_total, clusters_training, clusters_test


import pandas as pd
from rdkit.Chem import Descriptors
from rdkit.Chem import rdMolDescriptors
from rdkit.DataStructs import cDataStructs

def physicochemical_descriptors(file):
	NumHAcceptors = {}
	NumHDonors = {}
	ExactMolWt = {}
	NumRotatableBonds = {}
	FractionCSP3 = {}
	TPSA = {}
	LogP_WildmanCrippen = {}

	for i in file['molecules']:
		NumHAcceptors[i.GetProp('_Name')] = Descriptors.NumHAcceptors(i) # CALC: number of hydrogen acceptors
		NumHDonors[i.GetProp('_Name')] = Descriptors.NumHDonors(i) # CALC: number of hydrogen donors
		ExactMolWt[i.GetProp('_Name')] = Descriptors.ExactMolWt(i) # CALC: molecular weight
		NumRotatableBonds[i.GetProp('_Name')] = Descriptors.NumRotatableBonds(i) # CALC: number of rotatable bounds
		FractionCSP3[i.GetProp('_Name')] = Descriptors.FractionCSP3(i) # CALC: sp3 carbon fraction
		TPSA[i.GetProp('_Name')] = Descriptors.TPSA(i) # CALC: TPSA
		LogP_WildmanCrippen[i.GetProp('_Name')] = Descriptors.MolLogP(i) # CALC: WildmanCrippen Log P

	file['NumHAcceptors'] = pd.Series(NumHAcceptors)
	file['NumHDonors'] = pd.Series(NumHDonors)
	file['ExactMolWt'] = pd.Series(ExactMolWt)
	file['NumRotatableBonds'] = pd.Series(NumRotatableBonds)
	file['FractionCSP3'] = pd.Series(FractionCSP3)
	file['TPSA'] = pd.Series(TPSA)
	file['LogP_WildmanCrippen'] = pd.Series(LogP_WildmanCrippen)
	return file

def atompairs_fingerprint(file):
	FP_dict = {}
	for i in file['molecules']:
		FPasBitVect = rdMolDescriptors.GetHashedAtomPairFingerprintAsBitVect(i, nBits=1024) # Determine the fingerprint of the molecule using AtomPairs Fingerprint.
		FPasBinary = cDataStructs.BitVectToText(FPasBitVect) # Convert fingerprint to text.
		BinaryAsList = list(FPasBinary) # Each bit of fingerprint will be a string item of a list.
		BinaryAsNewList = []
		for a in BinaryAsList:
			b = int(a) # Each bit of fingerprint will be a int item of a list.
			BinaryAsNewList.append(b)
		FP_dict[i.GetProp('_Name')] = BinaryAsNewList # Create a "molecule name":fingerprint dictionary.

	file['AtomPairs'] = pd.Series(FP_dict) # Add the fingerprint to the dataframe.
	return file



from rdkit.Chem import AllChem

def output_mols(dataframe, FileOutput): # Add, for each molecule, the values of the calculated properties, the identifications of each cluster and which set the molecule belongs to.
	# Creation of dictionaries with the name of the molecule:
	dict_molecules = dict(dataframe['molecules']) # name:molecule
	dict_set = dict(dataframe['set']) # name:set
	dict_NumHAcceptors = dict(dataframe['NumHAcceptors']) # name:HBA
	dict_NumHDonors = dict(dataframe['NumHDonors']) # name:HBD
	dict_ExactMolWt = dict(dataframe['ExactMolWt']) # name:Molecular Weight
	dict_NumRotatableBonds = dict(dataframe['NumRotatableBonds']) # name:nRotBounds
	dict_FractionCSP3 = dict(dataframe['FractionCSP3']) # name:Fsp3
	dict_TPSA = dict(dataframe['TPSA']) # name:TPSA
	dict_LogP_WildmanCrippen = dict(dataframe['LogP_WildmanCrippen']) # name:logP
	dict_Cluster_General = dict(dataframe['Cluster_General']) # name:general_cluster
	dict_Cluster_Biological = dict(dataframe['Cluster_Biological']) # name:biological_cluster
	dict_Cluster_Physicochemical = dict(dataframe['Cluster_Physicochemical']) # name:physicochemical_cluster
	dict_Cluster_Structural = dict(dataframe['Cluster_Structural']) # name:strucutral_cluster

	# Add the properties to molecules from the name:molecule dictionary:
	for i in dict_molecules.keys():
		dict_molecules[i].SetProp('Set', str(dict_set[i]))
		dict_molecules[i].SetProp('HBA count', str(dict_NumHAcceptors[i]))
		dict_molecules[i].SetProp('HBD count', str(dict_NumHDonors[i]))
		dict_molecules[i].SetProp('ExactMolWt', str(dict_ExactMolWt[i]))
		dict_molecules[i].SetProp('NumRotatableBounds', str(dict_NumRotatableBonds[i]))
		dict_molecules[i].SetProp('FractionCsp3', str(dict_FractionCSP3[i]))
		dict_molecules[i].SetProp('TPSA', str(dict_TPSA[i]))
		dict_molecules[i].SetProp('LogP_WildmanCrippen', str(dict_LogP_WildmanCrippen[i]))
		dict_molecules[i].SetProp('General_Cluster', str(dict_Cluster_General[i]))
		dict_molecules[i].SetProp('Biological_Cluster', str(dict_Cluster_Biological[i]))
		dict_molecules[i].SetProp('Physicochemical_Cluster', str(dict_Cluster_Physicochemical[i]))
		dict_molecules[i].SetProp('Structural_Cluster', str(dict_Cluster_Structural[i]))

	# Create a dict_values with just the molecules (the name of the molecule is already part of molecule object):
	output_mols = dict_molecules.values()

	# Create output ".sdf" file:
	with AllChem.SDWriter(FileOutput) as w:
		for m in output_mols:
			w.write(m)


from MASSA_Algorithm import MASSAlogos
from MASSA_Algorithm import MASSAargs
from MASSA_Algorithm import MASSAod
from MASSA_Algorithm import MASSAopen_files
from MASSA_Algorithm import MASSAextraction
from MASSA_Algorithm import MASSAdescriptors
from MASSA_Algorithm import MASSApreparation
from MASSA_Algorithm import MASSAcluster
from MASSA_Algorithm import MASSAsplit
from MASSA_Algorithm import MASSAmoloutput

def returns_zero(total, test):
	# It evaluates if the distribution is not adequate = the iterated cluster has a percentage greater than 0.5% in the complete data set, but a percentage lower than 0.5% in the test set. This situation is unlikely to occur, as Kmodes were used precisely to avoid this.
	definer = False
	for i in total.keys():
		if (total[i] > 0.5) and (test[i] <= 0.5):
			definer = True # Definer = True (Distribution was not done properly).
	return definer

def main(): # Main subroutine, allows the program to run directly from the command line after installed via pip.
	## Initializing from the command line:
	MASSAlogos.initial_print() # Print the program logo.
	FileInput, FileOutput, directoryFileOutput, extension_type, dendrogram_Xfont_size, barplot_Xfont_size, training_percent, test_percent, numberBioAct, BioActAsArgs, nPCS, svd_parameter = MASSAargs.capture_args() # It captures command line arguments.
	print('Initializing, wait...\n')


	## Initial file management:
	MASSAod.output_directory(directoryFileOutput) # It creates the output directories.
	mols = MASSAopen_files.read_molecules(FileInput) # Read molecules.
	MASSAopen_files.error_check(mols) # Alert of RDKit fail to read.
	sdf_property_names = MASSAopen_files.get_sdf_property_names(mols) # Extracting the property names from the ".sdf" input file.
	molsH = MASSAopen_files.hydrogen_add(mols) # Structure 3D management - It adds hydrogens keeping 3D coordenates.


	## Extraction properties from ".sdf":
	names, dataframe = MASSAextraction.name_extraction(molsH) # It extracts the names of the molecules and creates a name:molecule dictionary and a dataframe.
	biological_activity = MASSAextraction.the_biological_handler(sdf_property_names, numberBioAct, BioActAsArgs) # It defines a list of what biological activities are being extracted.
	dataframe = MASSAextraction.list_activities(dataframe, biological_activity) # It adds the biological activities to the dataframe.


	## Get fingeprint and other descriptors:
	dataframe = MASSAdescriptors.physicochemical_descriptors(dataframe) # Get physicochemical descriptors.
	dataframe = MASSAdescriptors.atompairs_fingerprint(dataframe) # Get AtomPairs fingerprint.


	## Normalizes physicochemical and biological properties and creates matrices for the three domains:
	bio_matrix, PhCh_matrix, FP_matrix = MASSApreparation.normalizer_or_matrix(dataframe, biological_activity)


	## PCA:
	bio_PCA = MASSApreparation.pca_maker(bio_matrix, nPCS, svd_parameter) # PCA for the biological domain.
	PhCh_PCA = MASSApreparation.pca_maker(PhCh_matrix, nPCS, svd_parameter) # PCA for the physicochemical domain.
	FP_PCA = MASSApreparation.pca_maker(FP_matrix, nPCS, svd_parameter) # PCA for the structural domain.


	## First clustering (HCA):
	leaves_cluster_bio, bioHCA, linkage_bio, CutOff_bio = MASSAcluster.hca_clusters(bio_PCA, names, 'bio', directoryFileOutput, extension_type) # It performs HCA clustering without generating the dendrogram for the biological domain.
	leaves_cluster_phch, phchHCA, linkage_phch, CutOff_phch = MASSAcluster.hca_clusters(PhCh_PCA, names, 'PhCh', directoryFileOutput, extension_type) # It performs HCA clustering without generating the dendrogram for the physicochemical domain.
	leaves_cluster_fp, fpHCA, linkage_fp, CutOff_fp = MASSAcluster.hca_clusters(FP_PCA, names, 'FP', directoryFileOutput, extension_type) # It performs HCA clustering without generating the dendrogram for the structural domain.

	dataframe = MASSApreparation.organize_hca(dataframe, bioHCA, 'bio') # It adds the biological cluster identification to the spreadsheet.
	dataframe = MASSApreparation.organize_hca(dataframe, phchHCA, 'PhCh') # It adds the physicochemical cluster identification to the spreadsheet.
	dataframe = MASSApreparation.organize_hca(dataframe, fpHCA, 'FP') # It adds the structural cluster identification to the spreadsheet.


	## Second clustering (Kmodes):
	matrix_for_kmodes = MASSApreparation.organize_for_kmodes(dataframe) # It creates a matrix with cluster identifications for each of the three domains, in order to prepare for Kmodes.
	allHCA = MASSAcluster.kmodes_clusters(matrix_for_kmodes, names) # It performs Kmodes clustering for the general domain.
	dataframe = MASSApreparation.organize_hca(dataframe, allHCA, 'all') # It adds the general cluster identification to the spreadsheet.


	## Split into training, test:
	dataframe, test_molecules = MASSAsplit.split_train_test_sets(dataframe, training_percent, test_percent)


	## Bar plot of frequencies (Calculates the percentages of molecules in each cluster for each dataset and generates a bar graph for each domain):
	bio_total, bio_training, bio_test = MASSAsplit.freq_clusters(dataframe, directoryFileOutput, extension_type, 'Cluster_Biological', barplot_Xfont_size) # Biological Bar Plot
	PhCh_total, PhCh_training, PhCh_test = MASSAsplit.freq_clusters(dataframe, directoryFileOutput, extension_type, 'Cluster_Physicochemical', barplot_Xfont_size) # Physicochemical Bar Plot
	FP_total, FP_training, FP_test = MASSAsplit.freq_clusters(dataframe, directoryFileOutput, extension_type, 'Cluster_Structural', barplot_Xfont_size) # Structural Bar Plot
	all_total, all_training, all_test = MASSAsplit.freq_clusters(dataframe, directoryFileOutput, extension_type, 'Cluster_General', barplot_Xfont_size) # General Bar Plot


	## Verifying percentages:
	bio_ok = returns_zero(bio_total, bio_test) # For biological domain.
	PhCh_ok = returns_zero(PhCh_total, PhCh_test) # For physicochemical domain.
	FP_ok = returns_zero(FP_total, FP_test) # For structural domain.
	ok = [bio_ok, PhCh_ok, FP_ok]
	max_iters = 0
		
	# Redo the distribution in case of errors (up to 5 times):
	while (True in ok) and (max_iters < 5):
		## Split into training, test:
		dataframe, test_molecules = MASSAsplit.split_train_test_sets(dataframe, training_percent, test_percent)

		## Bar plot of frequencies (Calculates the percentages of molecules in each cluster for each dataset and generates a bar graph for each domain):
		bio_total, bio_training, bio_test = MASSAsplit.freq_clusters(dataframe, directoryFileOutput, extension_type, 'Cluster_Biological', barplot_Xfont_size) # Biological Bar Plot
		PhCh_total, PhCh_training, PhCh_test = MASSAsplit.freq_clusters(dataframe, directoryFileOutput, extension_type, 'Cluster_Physicochemical', barplot_Xfont_size) # Physicochemical Bar Plot
		FP_total, FP_training, FP_test = MASSAsplit.freq_clusters(dataframe, directoryFileOutput, extension_type, 'Cluster_Structural', barplot_Xfont_size) # Structural Bar Plot
		all_total, all_training, all_test = MASSAsplit.freq_clusters(dataframe, directoryFileOutput, extension_type, 'Cluster_General', barplot_Xfont_size) # General Bar Plot

		## Verifying percentages:
		bio_ok = returns_zero(bio_total, bio_test) # For biological domain.
		PhCh_ok = returns_zero(PhCh_total, PhCh_test) # For physicochemical domain.
		FP_ok = returns_zero(FP_total, FP_test) # For structural domain.
		ok = [bio_ok, PhCh_ok, FP_ok]
		max_iters += 1


	## Plot HCAs:
	print('\nGenerating dendrogram images. Please wait...')
	MASSAcluster.hca_plot(linkage_bio, names, leaves_cluster_bio, CutOff_bio, 'bio', directoryFileOutput, extension_type, dendrogram_Xfont_size, test_molecules) #Bio_Plot: Plot the HCA dendrogram
	MASSAcluster.hca_plot(linkage_phch, names, leaves_cluster_phch, CutOff_phch, 'PhCh', directoryFileOutput, extension_type, dendrogram_Xfont_size, test_molecules) #PhCh_Plot: Plot the HCA dendrogram
	MASSAcluster.hca_plot(linkage_fp, names, leaves_cluster_fp, CutOff_fp, 'FP', directoryFileOutput, extension_type, dendrogram_Xfont_size, test_molecules) #FP_Plot: Plot the HCA dendrogram


	## Output management:
	MASSAmoloutput.output_mols(dataframe, FileOutput) # It adds, for each molecule, the values of the calculated properties, the identifications of each cluster and which set the molecule belongs to.
	print('Completed')


import pandas as pd
from rdkit.Chem import AllChem

def name_extraction(file):
	names = []
	dict_names = {}
	for i in file:
		try:
			names.append(i.GetProp('_Name')) # Get name of molecules to a list
			dict_names[i.GetProp('_Name')] = i # Get name of molecules to a dict
		except:
			names.append(None)
	if None in names: # If any molecule is unnamed it returns an error.
		print('ERROR: Couldn\'t get any name from a molecule. Adding names before running the script is essential. Check which molecule is in the ordered list below: \n', names)
		print('Closing...')
		exit()
	else:
		dataframe = pd.DataFrame({'molecules': pd.Series(dict_names)})
	return names, dataframe

def the_biological_handler(sdf_property_names, numberBioAct, BioActAsArgs):
	if BioActAsArgs == None: # If no biological activity name is passed as a command line argument.
		if (numberBioAct == len(sdf_property_names)): # If the number of biological activities equals the number of properties in the file (not counting the name), the only available properties will automatically be identified as the biological activities.
			name_biological_list = sdf_property_names.copy()
			return name_biological_list
		elif (numberBioAct > len(sdf_property_names)): # If the number of biological activities is greater than the number of properties in the file (not counting the name), returns an error.
			print('ERROR: The number of biological activities defined is greater than the number of properties in the .sdf file.')
			exit()
		else: # If no biological activity name was passed as a command line argument and the number of biological activities is less than the number of properties: Prompts user iteration to define the names of biological activities.
			name_biological_list = []
			print('Found the following properties in the input file: ' + str(sdf_property_names) + '.')
			for i in range(0,numberBioAct):
				a = (input('Enter the name of the ' + str(i+1) + 'st/nd/rd/th biological activity:  '))
				max_iterations = 0
				while (a not in sdf_property_names) and (max_iterations < 3): # If any property is typed incorrectly, the user will have 3 attempts to type correctly.
					print('ERROR: The biological property name entered does not match any of the properties found. Try again or exit the program from the command line using Ctrl+C.')
					a = (input('Enter the name of the ' + str(i+1) + 'st/nd/rd/th biological activity:  '))
					max_iterations += 1
				if a not in sdf_property_names: # If even after 3 attempts the name of the activity is still incorrect, the program will close.
					print('ERROR: The name of the biological activity was not typed correctly.')
					exit()
				name_biological_list.append(a)
			return name_biological_list

	else: # If biological activity names are passed as a command-line argument.
		if (numberBioAct == len(BioActAsArgs)): # The number of biological activities defined by the command line must be equal to the number of biological activities passed by the name.
			name_biological_list = BioActAsArgs.copy()
			for i in name_biological_list:
				if i not in sdf_property_names: # If the properties entered by the command line are typed incorrectly the program will close.
					print('ERROR: The name of the biological activity was not typed correctly.')
					exit()
			return name_biological_list
		else:
			print('ERROR: The number of biological activities passed as list \'-s\' is different from the number of biological activities passed as int \'b\'.')
			exit()

def list_activities(file, name_biological_activity):
	for bioACname in name_biological_activity: # For each biological activity:
		bio_dict = {}
		for i in file['molecules']: # For each molecule in the dataframe:
			try:
				bio_name = i.GetProp('_Name') # Capture the name of the molecule.
				bio_value = (float(i.GetProp(bioACname))) # Capture the value of the biological activity.
				bio_dict[bio_name] = bio_value # Add the "molecule name":"value" pair to the dictionary.
			except: # If any molecule does not have the biological activity in question it will add NoneType in place of the value.
				bio_name = i.GetProp('_Name')
				bio_value = None
				bio_dict[bio_name] = bio_value
		if(set(list(bio_dict.values()))) == {None}: # If no molecule has the biological property in question, it closes the program reporting the problem found.
			print('ERROR: The chosen property does not contain data or the name of the biological activity was not typed correctly.')
			exit()
		elif None in bio_dict.values(): # If any molecule has no biological activity it will return an error informing this and will print the biological activity in question and the dictionary "molecule name":"value".
			print('ERROR: Some molecule has no biological activity value.')
			print('Biological activity:', bioACname)
			print('Molecules:', bio_dict)
			exit()
		file[bioACname] = pd.Series(bio_dict)
	return file


import os

def output_directory(directoryFileOutput): # Create the output directories.
	try:
		os.makedirs(directoryFileOutput+'Images')
	except FileExistsError:
		pass
	try:
		os.makedirs(directoryFileOutput+'Images/Distance_HCA_Images')
	except FileExistsError:
		pass
	try:
		os.makedirs(directoryFileOutput+'Images/Dendrogram_HCA_Images')
	except FileExistsError:
		pass
	try:
		os.makedirs(directoryFileOutput+'Images/Distribution')
	except FileExistsError:
		pass


from rdkit.Chem import AllChem

def read_molecules(file):
	mols = AllChem.SDMolSupplier(file) # Open the input file.
	return mols

def error_check(file):
	for i in file:
		if i == None: # If any molecule is not able to be read: 
			print('ERROR: Unable to read a molecule') # Print an error.
			exit()
		else:
			pass

def hydrogen_add(file):
	molsH = []
	for i in file:
		molsH.append(AllChem.AddHs(i, addCoords = True)) # Add hydrogens keeping 3D coordenates.
	return molsH

def get_sdf_property_names(file): # Extracting the property names from the ".sdf" input file.
	sdf_property_names = []
	for i in file:
		for a in list(i.GetPropNames()):
			sdf_property_names.append(a)
	sdf_property_names = list(set(sdf_property_names))
	if 'Name' in sdf_property_names: # Drop the molecule name from the property list.
		sdf_property_names.remove('Name')
	return sdf_property_names


import setuptools

with open('README.md', 'r') as readme:
	long_description = readme.read()

setuptools.setup(
	name='MASSA Algorithm',
	version=0.1,
	description='MASSA Algorithm is a Python package to separate data sets of molecules into training and test sets, considering the diversity of structural, physicochemical and biological characteristics of these molecules.',
	long_description=long_description,
	long_description_content_type='text/markdown',
	author='Gabriel Corrêa Veríssimo',
	author_email='gcverissimo@outlook.com',
	url='https://github.com/gcverissimo/MASSA_Algorithm',
	license='AGPLv3',
	packages= setuptools.find_packages(),
	entry_points={'console_scripts':['MASSA_Algorithm = MASSA_Algorithm.MASSA:main']},
	install_requires=['numpy', 'pandas', 'matplotlib >= 3.2', 'scipy >= 1.6', 'scikit-learn >= 0.24', 'kmodes >= 0.10'],
	python_requires='>=3.7, <4',
	keywords=['chemoinformatics', 'training', 'test', 'training-test', 'dataset preparation', 'data set preparation', 'rdkit'],
	classifiers=['Programming Language :: Python :: 3',
	'Programming Language :: Python :: 3.7',
	'Programming Language :: Python :: 3.8',
	'Programming Language :: Python :: 3.9',
	'License :: OSI Approved :: GNU Affero General Public License v3',
	'Operating System :: OS Independent',
	'Topic :: Scientific/Engineering :: Chemistry'])


import re
import os
import argparse
import numpy as np

def capture_args():
	## Capturing arguments from command line:
	parser = argparse.ArgumentParser(prog='MASSA_Algorithm', epilog='For more information visit: https://github.com/gcverissimo/MASSA_Algorithm', description='Help for MASSA Algorithm: Molecular data set sampling  —  Training-Test Separation', usage='%(prog)s [-h] [-i <input_file_name>] [-o <output_file_name] [-t <extension_for_images>] [-p <percentage_of_molecules_in_training_set>] [-b <number_of_biological_activities_in_input_file>], [-s <activity1>,<activity2>,<activityN>], [-n <number_of_principal_components>], [-v <svd_parameter>], [-d <dendrogram_font_size_on_Xaxis>], [-x <frequency_bar_plot_font_size_on_Xaxis>]')
	required = parser.add_argument_group('required arguments')
	required.add_argument('-i','--input', type=str, help='Input the file containing the dataset or its file path. Recommended to use an .sdf file to avoid errors.', required=True)
	required.add_argument('-o','--output', type=str, help='Enter the output file name or file path. Recommended to use an .sdf file to avoid errors. Image files will be saved to a folder within the same directory as the output file.', required=True)
	optional = parser.add_argument_group('optional arguments')
	optional.add_argument('-t','--image_type', type=str, default='png', help='Extension of the image files that will be generated. Suggested = png or svg. Default = png.')
	optional.add_argument('-d','--dendrogram_Xfont_size', type=int, default=5, help='Sets the font size on the x-axis of the dendrogram (molecule labels). Default = 5.')
	optional.add_argument('-x','--barplot_Xfont_size', type=int, default=9, help='Sets the font size on the x-axis of the bar plot (cluster labels). Default = 9.')
	optional.add_argument('-p','--percentage_of_training', type=float, choices=[round(i,2) for i in list(np.arange(0.00, 1.00, 0.01))], metavar='PERCENTAGE_OF_TRAINING', default=0.8, help='Percentage of molecules in training set. Must be a number from 0 to 1. Default = 0.8.')
	optional.add_argument('-b','--number_of_biological', type=int, default=1, help='Number of biological activities that will be used to separate the set into training and test. Default = 1')
	optional.add_argument('-s','--the_biological_activities', type=str, default=None, help='Enter a list with the names of biological activities separated by commas and no spaces. Example: pIC50,pMIC.')
	optional.add_argument('-n','--number_of_PCs', type=float, metavar='NUMBER_OF_PRINCIPAL_COMPONENTS', default=0.85, help='Defines the number of principal components to reduce the dimensionality of variables related to biological, physicochemical and structural domains. If the value is a decimal between 0 and 1, the number of principal components is what justifies for (<input number> * 100) percent of the variance. If the value is greater than 1, the number of PCs will be exactly the input integer, but PAY ATTENTION: 1st) If the number of PCs is an integer and equal to or greater than the number of physicochemical properties (7), the PCA step will be bypassed for this domain. 2nd) The same for the biological domain. 3rd) If the number of biological activities is less than 3, the PCA step will be bypassed for this domain. Default = 0.85.')
	optional.add_argument('-v','--svd_solver_for_PCA', type=str, default='auto', help='See the sklearn.decomposition.PCA topic on https://scikit-learn.org/ for more info. Default = auto.')
	args = parser.parse_args()

	## Definition of arguments:
	FileInput = args.input # It captures the input name and directory+name.
	undefinedFileOutput = args.output # It captures the output name or directory+name.
	extension_type = args.image_type # It captures the file extension for images.
	dendrogram_Xfont_size = args.dendrogram_Xfont_size # It captures the font size for the x-axis of the dendrogram.
	barplot_Xfont_size = args.barplot_Xfont_size # It captures the font size for the x-axis of the bar plot.
	training_percent = args.percentage_of_training # It captures the percentage for training set.
	test_percent = 1 - training_percent # Calculates the percentage of molecules in the test set.
	numberBioAct = args.number_of_biological # It captures the number of biological activities.
	inputnumberPCs = args.number_of_PCs # It captures the input number of PCs.
	svd_parameter = args.svd_solver_for_PCA # It captures the svd_solver parameter.

	if args.the_biological_activities != None:
		BioActAsArgs = args.the_biological_activities.split(',') # It creates a list with the names of the biological activities to be searched for.
	else:
		BioActAsArgs = None 

	if ('\\' in undefinedFileOutput) or ('/' in undefinedFileOutput): # If a directory path is passed along with the filename it copies only the directory path.
		outputsplitted = re.split('\\\\|/', undefinedFileOutput)
		outputsplitted[-1] = ''
		directoryFileOutput = '\\'.join(outputsplitted)
		del outputsplitted
	else: # If only the filename is passed, the directory path is taken by os.getcwd().
		directoryFileOutput = str(os.getcwd()) + '\\'
	
	if (inputnumberPCs > 1):
		nPCS = int(inputnumberPCs)
	elif (inputnumberPCs > 0) and (inputnumberPCs <= 1):
		nPCS = float(inputnumberPCs)
	else:
		print(' ERROR: The defined number of main components is invalid.')
		exit()

	return FileInput, undefinedFileOutput, directoryFileOutput, extension_type, dendrogram_Xfont_size, barplot_Xfont_size, training_percent, test_percent, numberBioAct, BioActAsArgs, nPCS, svd_parameter